import sys
from traceback import linecache


def default_printer(locals_):
    return ', '.join([f'{k}={v}' for k, v in locals_.items()])


def names_printer(locals_, names):
    return ', '.join([f'{k}={v}' for k, v in locals_.items() if k in names])


def multiline_printer(locals_, width=80):
    s = ' ' * width
    return '\n' + '\n'.join([f'{s}{k}={v}' for k, v in locals_.items()]) + '\n'


class trace_context:
    def __init__(self, name, printer=None, width=None):
        self.name = name
        self.printer = printer or default_printer
        self.width = width or 80
        self.old_trace = None

    def __enter__(self):
        print(f'\nEntering function: {self.name}')
        self.old_trace = sys.gettrace()
        sys.settrace(self.trace_calls)

    def __exit__(self, *args, **kwargs):
        sys.settrace(self.old_trace)

    def trace_calls(self, frame, event, arg):
        if event != 'call' or frame.f_code.co_name != self.name:
            return None
        return self.trace_lines

    def trace_lines(self, frame, event, arg):
        if event not in ['line', 'return']:
            return
        line = linecache.getline(frame.f_code.co_filename, frame.f_lineno, frame.f_globals).rstrip()
        print(f'{line.ljust(self.width)}locals: {self.printer(frame.f_locals)}')
        if event == 'return':
            prev_name = f'{self.old_trace.__self__.name}()' if self.old_trace else 'None'
            print(f'Returning from: {self.name}() to: {prev_name}\n')


def tracelogger(*args, **kwargs):
    if len(set(kwargs.keys()) - {'printer', 'width'}) > 0:
        raise ValueError('Unknown parameter(s)')
    if len(kwargs) == 0:
        func = args[0]

        def decorated_func(*func_args, **func_kwargs):
            with trace_context(func.__name__):
                return_value = func(*func_args, **func_kwargs)
            return return_value

        return decorated_func
    else:

        def wrapper(func):
            def decorated_func(*func_args, **func_kwargs):
                printer = kwargs.get('printer', args[0] if args else None)
                width = kwargs.get('width', args[1] if args and len(args) > 1 else None)
                with trace_context(func.__name__, printer=printer, width=width):
                    return_value = func(*func_args, **func_kwargs)
                return return_value

            return decorated_func

        return wrapper



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['tracelogger']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'tracelogger',
    'version': '0.1.2',
    'description': '',
    'long_description': "# Tracelogger\n\nTag your functions with the `@tracelogger` decorator to be logged line-by-line.\n\nOptionally provide a function as the `printer` argument that takes the local variable dictionary and returns a string. This allows the customisation of which variables and how you want to print at each line. See the example with the provided `names_printer` function that only prints a selected set of variables.\n\nThe second `width` parameter determines the maximum character number for each line of code defaulting to `80`.\n\n```\nfrom tracelogger import tracelogger, names_printer\n\n\n@tracelogger(width=30)\ndef second_test_function(x):\n    y = x**2\n    return y\n\n\n@tracelogger(printer=lambda locals_: names_printer(locals_=locals_, names=['k', 'b']), width=30)\ndef first_test_function(a, b):\n    c = a + b\n    for k in range(5):\n        if k % 2 == 0:\n            b += k\n        else:\n            c += k\n            second_test_function(c)\n    return a, b, c\n\na, b, c = first_test_function(a=10, b=20)\n```\n\nThis will print the following:\n\n```\nEntering function: first_test_function\n    c = a + b                            locals: b=20\n    for k in range(5):                   locals: b=20\n        if k % 2 == 0:                   locals: b=20, k=0\n            b += k                       locals: b=20, k=0\n    for k in range(5):                   locals: b=20, k=0\n        if k % 2 == 0:                   locals: b=20, k=1\n            c += k                       locals: b=20, k=1\n            second_test_function(c)      locals: b=20, k=1\n\nEntering function: second_test_function\n    y = x**2                             locals: x=31\n    return y                             locals: x=31, y=961\n    return y                             locals: x=31, y=961\nReturning from: second_test_function() to: first_test_function()\n\n    for k in range(5):                   locals: b=20, k=1\n        if k % 2 == 0:                   locals: b=20, k=2\n            b += k                       locals: b=20, k=2\n    for k in range(5):                   locals: b=22, k=2\n        if k % 2 == 0:                   locals: b=22, k=3\n            c += k                       locals: b=22, k=3\n            second_test_function(c)      locals: b=22, k=3\n\nEntering function: second_test_function\n    y = x**2                             locals: x=34\n    return y                             locals: x=34, y=1156\n    return y                             locals: x=34, y=1156\nReturning from: second_test_function() to: first_test_function()\n\n    for k in range(5):                   locals: b=22, k=3\n        if k % 2 == 0:                   locals: b=22, k=4\n            b += k                       locals: b=22, k=4\n    for k in range(5):                   locals: b=26, k=4\n    return a, b, c                       locals: b=26, k=4\n    return a, b, c                       locals: b=26, k=4\nReturning from: first_test_function() to: None\n```\n\nJoin the [Code Quality for Data Science (CQ4DS) Discord channel](https://discord.com/invite/8uUZNMCad2) for feedback.\n\nI used the following StackOverflow threads as sources, many thanks to their authors:\n\n[https://stackoverflow.com/questions/32163436/python-decorator-for-printing-every-line-executed-by-a-function](https://stackoverflow.com/questions/32163436/python-decorator-for-printing-every-line-executed-by-a-function)\n\n[https://stackoverflow.com/questions/22362940/inspect-code-of-next-line-in-python](https://stackoverflow.com/questions/22362940/inspect-code-of-next-line-in-python)\n",
    'author': 'Laszlo Sragner',
    'author_email': 'sragner@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)
